@@ -5,14 +5,16 @@
 #  ALL RIGHTS RESERVED.
 #
 #  Authors: Vasya Svintsov <v.svintsov@techokert.ru>
+import copy
 from contextlib import contextmanager, asynccontextmanager
 from dataclasses import dataclass
 from threading import Lock
-from typing import Sequence, Any, Iterator, Callable, Awaitable
+from typing import Sequence, Any, Iterator, Callable, Awaitable, Iterable
 
 from async_tools import acall
 from more_itertools import first
-from prometheus_client import CollectorRegistry, Counter, Gauge, Summary, Histogram
+from prometheus_client import Counter, Gauge, Summary, Histogram, Metric, generate_latest
+from prometheus_client.metrics import MetricWrapperBase
 from prometheus_client.utils import INF
 
 from .tools.call import call
@@ -27,7 +29,10 @@ class BaseMetricsRegistry:
     def __init__(self, config: Config) -> None:
         self.config = config
         self._lock = Lock()
-        self._registry = CollectorRegistry(auto_describe=True)
+        self._name_to_collector: dict[tuple[str, ...], MetricWrapperBase] = {}
+
+    def render_metrics(self) -> bytes:
+        return generate_latest(type("registry", (), {'collect': self._collect})())
 
     def counter(self, name: str, labels: dict[str, Any] | None = None, doc: str = '') -> Counter:
         return self._ensure_collector(Counter, name, labels, doc)
@@ -53,22 +58,31 @@ class BaseMetricsRegistry:
             labels = labels or {}
             label_keys = set(labels.keys())
 
-            collector = self._registry._names_to_collectors.get(name)  # bad library interface
+            collector_name = (name, *label_keys)
+            collector = self._name_to_collector.get(collector_name)
             if not collector:
-                collector = required_type(name, doc, label_keys, registry=self._registry, **kwargs)
+                collector = required_type(name, doc, label_keys, registry=None, **kwargs)
+                self._name_to_collector[collector_name] = collector
             elif not isinstance(collector, required_type):
                 raise ValueError(f"The name '{name}' is already taken by the {type(collector)}")
-            elif mismatched_labels := set(collector._labelnames) ^ label_keys:
-                raise ValueError(f"The existing '{repr(collector)}' already has labels: {collector._labelnames}. "
-                                 f"Required labels: {label_keys}, mismatched labels: {mismatched_labels}")
 
             if labels:
                 collector = collector.labels(**labels)
 
         return collector
 
+    def _collect(self) -> Iterable[Metric]:
+        with self._lock:
+            collectors = copy.copy(list(self._name_to_collector.values()))
+        for collector in collectors:
+            yield from collector.collect()
+
 
 class MetricsRegistry(BaseMetricsRegistry):
+    @dataclass
+    class Tracker:
+        labels: dict[str, Any]
+
     @contextmanager
     def track_progress(self, name: str, labels: dict[str, Any] | None = None, doc: str = '') -> Iterator[None]:
         with self.gauge(name, labels, doc).track_inprogress():
@@ -92,10 +106,10 @@ class MetricsRegistry(BaseMetricsRegistry):
             *,
             except_labels: dict[str, Any] | None = None,
             missing_label_value: Callable[[], Any] | Any | None = None
-    ) -> Iterator[None]:
+    ) -> Iterator[Tracker]:
         missing_label_values = []
-        with self._track(missing_label_values, name, labels, except_labels):
-            yield
+        with self._track(missing_label_values, name, labels, except_labels) as tracker:
+            yield tracker
             missing_label_values.append(call(missing_label_value))
 
     @asynccontextmanager
@@ -106,10 +120,10 @@ class MetricsRegistry(BaseMetricsRegistry):
             *,
             except_labels: dict[str, Any] | None = None,
             missing_label_value: Callable[[], Awaitable[Any] | Any] | Any | None = None
-    ) -> Iterator[None]:
+    ) -> Iterator[Tracker]:
         missing_label_values = []
-        with self._track(missing_label_values, name, labels, except_labels):
-            yield
+        with self._track(missing_label_values, name, labels, except_labels) as tracker:
+            yield tracker
             missing_label_values.append(await acall(missing_label_value))
 
     @contextmanager
@@ -119,14 +133,14 @@ class MetricsRegistry(BaseMetricsRegistry):
             name: str,
             labels: dict[str, Any] | None = None,
             except_labels: dict[str, Any] | None = None
-    ) -> Iterator[None]:
+    ) -> Iterator[Tracker]:
         labels = labels or {}
         except_labels = except_labels or {}
         required_keys = labels.keys() | except_labels.keys()
         with self.track_progress(f'{name}__progress', labels):
             with self.track_time(f'{name}__spent_time', self.config.default_buckets, labels):
                 try:
-                    yield
+                    yield self.Tracker(labels)
                     labels |= {key: first(missing_label_values, None) for key in required_keys - labels.keys()}
                 except Exception:
                     labels |= except_labels
